import pytest
import ribasim
import ribasim_testmodels


# we can't call fixtures directly, so we keep separate versions
@pytest.fixture()
def basic() -> ribasim.Model:
    return ribasim_testmodels.basic_model()


@pytest.fixture()
def basic_arrow() -> ribasim.Model:
    return ribasim_testmodels.basic_arrow_model()


@pytest.fixture()
def basic_transient() -> ribasim.Model:
    return ribasim_testmodels.basic_transient_model()


@pytest.fixture()
def tabulated_rating_curve() -> ribasim.Model:
    return ribasim_testmodels.tabulated_rating_curve_model()


@pytest.fixture()
def backwater() -> ribasim.Model:
    return ribasim_testmodels.backwater_model()


@pytest.fixture()
def discrete_control_of_pid_control() -> ribasim.Model:
    return ribasim_testmodels.discrete_control_of_pid_control_model()


@pytest.fixture()
def level_setpoint_with_minmax() -> ribasim.Model:
    return ribasim_testmodels.level_setpoint_with_minmax_model()


@pytest.fixture()
def trivial() -> ribasim.Model:
    return ribasim_testmodels.trivial_model()



from collections.abc import Sequence
from typing import Any

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandera as pa
import shapely
from matplotlib.patches import Patch
from numpy.typing import NDArray
from pandera.typing import Series
from pandera.typing.geopandas import GeoSeries
from pydantic import field_validator

from ribasim.input_base import SpatialTableModel

__all__ = ("Node",)


class NodeSchema(pa.SchemaModel):
    node_id: Series[int]
    name: Series[str] = pa.Field(default="")
    node_type: Series[str] = pa.Field(default="")
    subnetwork_id: Series[pd.Int64Dtype] = pa.Field(
        default=pd.NA, nullable=True, coerce=True
    )
    geometry: GeoSeries[Any] = pa.Field(default=None, nullable=True)

    class Config:
        add_missing_columns = True
        coerce = True


class Node(SpatialTableModel[NodeSchema]):
    """The Ribasim nodes as Point geometries."""

    # TODO: Remove as soon as add api has been merged
    @field_validator("df", mode="before")
    @classmethod
    def add_node_id_column(cls, df: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        if "node_id" not in df.columns:
            df.insert(0, "node_id", df.index)
        return df

    @staticmethod
    def node_ids_and_types(*nodes):
        # TODO Not sure if this staticmethod belongs here
        data_types = {"node_id": int, "node_type": str}
        node_type = pd.DataFrame(
            {col: pd.Series(dtype=dtype) for col, dtype in data_types.items()}
        )

        for node in nodes:
            if not node:
                continue

            ids, types = node.node_ids_and_types()
            node_type_table = pd.DataFrame(
                data={
                    "node_id": ids,
                    "node_type": types,
                }
            )
            node_type = node_type._append(node_type_table)

        node_type = node_type.drop_duplicates(subset="node_id")
        node_type = node_type.sort_values("node_id")

        node_id = node_type.node_id.tolist()
        node_type = node_type.node_type.tolist()

        return node_id, node_type

    def geometry_from_connectivity(
        self, from_id: Sequence[int], to_id: Sequence[int]
    ) -> NDArray[Any]:
        """
        Create edge shapely geometries from connectivities.

        Parameters
        ----------
        node : Ribasim.Node
        from_id : Sequence[int]
            First node of every edge.
        to_id : Sequence[int]
            Second node of every edge.

        Returns
        -------
        edge_geometry : np.ndarray
            Array of shapely LineStrings.
        """
        assert self.df is not None
        geometry = self.df["geometry"]
        from_points = shapely.get_coordinates(geometry.loc[from_id])
        to_points = shapely.get_coordinates(geometry.loc[to_id])
        n = len(from_points)
        vertices = np.empty((n * 2, 2), dtype=from_points.dtype)
        vertices[0::2, :] = from_points
        vertices[1::2, :] = to_points
        indices = np.repeat(np.arange(n), 2)
        return shapely.linestrings(coords=vertices, indices=indices)

    def connectivity_from_geometry(
        self, lines: NDArray[Any]
    ) -> tuple[NDArray[Any], NDArray[Any]]:
        """
        Derive from_node_id and to_node_id for every edge in lines. LineStrings
        may be used to connect multiple nodes in a sequence, but every linestring
        vertex must also a node.

        Parameters
        ----------
        node : Node
        lines : np.ndarray
            Array of shapely linestrings.

        Returns
        -------
        from_node_id : np.ndarray of int
        to_node_id : np.ndarray of int
        """
        assert self.df is not None
        node_index = self.df.index
        node_xy = shapely.get_coordinates(self.df.geometry.values)
        edge_xy = shapely.get_coordinates(lines)

        xy = np.vstack([node_xy, edge_xy])
        _, inverse = np.unique(xy, return_inverse=True, axis=0)
        _, index, inverse = np.unique(
            xy, return_index=True, return_inverse=True, axis=0
        )
        uniques_index = index[inverse]

        node_node_id, edge_node_id = np.split(uniques_index, [len(node_xy)])
        if not np.isin(edge_node_id, node_node_id).all():
            raise ValueError(
                "Edge lines contain coordinates that are not in the node layer. "
                "Please ensure all edges are snapped to nodes exactly."
            )

        edge_node_id = edge_node_id.reshape((-1, 2))
        from_id = node_index[edge_node_id[:, 0]].to_numpy()
        to_id = node_index[edge_node_id[:, 1]].to_numpy()
        return from_id, to_id

    def plot_allocation_networks(self, ax=None, zorder=None) -> Any:
        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")

        COLOR_SUBNETWORK = "black"
        COLOR_MAIN_NETWORK = "blue"
        ALPHA = 0.25

        contains_main_network = False
        contains_subnetworks = False
        assert self.df is not None

        for subnetwork_id, df_subnetwork in self.df.groupby("subnetwork_id"):
            if subnetwork_id is None:
                continue
            elif subnetwork_id == 1:
                contains_main_network = True
                color = COLOR_MAIN_NETWORK
            else:
                contains_subnetworks = True
                color = COLOR_SUBNETWORK

            hull = gpd.GeoDataFrame(
                geometry=[df_subnetwork.geometry.unary_union.convex_hull]
            )
            hull.plot(ax=ax, color=color, alpha=ALPHA, zorder=zorder)

        handles = []
        labels = []

        if contains_main_network:
            handles.append(Patch(facecolor=COLOR_MAIN_NETWORK, alpha=ALPHA))
            labels.append("Main network")
        if contains_subnetworks:
            handles.append(Patch(facecolor=COLOR_SUBNETWORK, alpha=ALPHA))
            labels.append("Subnetwork")

        return handles, labels

    def plot(self, ax=None, zorder=None) -> Any:
        """
        Plot the nodes. Each node type is given a separate marker.

        Parameters
        ----------
        ax : Optional
            The axis on which the nodes will be plotted.

        Returns
        -------
        None
        """
        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")

        MARKERS = {
            "Basin": "o",
            "FractionalFlow": "^",
            "LevelBoundary": "o",
            "LinearResistance": "^",
            "ManningResistance": "D",
            "TabulatedRatingCurve": "D",
            "Pump": "h",
            "Outlet": "h",
            "Terminal": "s",
            "FlowBoundary": "h",
            "DiscreteControl": "*",
            "PidControl": "x",
            "UserDemand": "s",
            "LevelDemand": "o",
            "": "o",
        }

        COLORS = {
            "Basin": "b",
            "FractionalFlow": "r",
            "LevelBoundary": "g",
            "LinearResistance": "g",
            "ManningResistance": "r",
            "TabulatedRatingCurve": "g",
            "Pump": "0.5",  # grayscale level
            "Outlet": "g",
            "Terminal": "m",
            "FlowBoundary": "m",
            "DiscreteControl": "k",
            "PidControl": "k",
            "UserDemand": "g",
            "LevelDemand": "k",
            "": "k",
        }
        assert self.df is not None

        for nodetype, df in self.df.groupby("node_type"):
            assert isinstance(nodetype, str)
            marker = MARKERS[nodetype]
            color = COLORS[nodetype]
            ax.scatter(
                df.geometry.x,
                df.geometry.y,
                marker=marker,
                color=color,
                zorder=zorder,
                label=nodetype,
            )

        assert self.df is not None
        geometry = self.df["geometry"]
        for text, xy in zip(self.df.index, np.column_stack((geometry.x, geometry.y))):
            ax.annotate(text=text, xy=xy, xytext=(2.0, 2.0), textcoords="offset points")

        return ax



import geopandas as gpd
import pytest
import shapely.geometry as sg
from pydantic import ValidationError
from ribasim.geometry.edge import Edge


@pytest.fixture(scope="session")
def edge() -> Edge:
    a = (0.0, 0.0)
    b = (0.0, 1.0)
    c = (0.2, 0.5)
    d = (1.0, 1.0)
    geometry = [sg.LineString([a, b, c]), sg.LineString([a, d])]
    df = gpd.GeoDataFrame(
        data={"from_node_id": [1, 1], "to_node_id": [2, 3]}, geometry=geometry
    )
    edge = Edge(df=df)
    return edge


def test_validation(edge):
    assert isinstance(edge, Edge)

    with pytest.raises(ValidationError):
        df = gpd.GeoDataFrame(
            data={"from_node_id": [1, 1], "to_node_id": [None, 3]},
            geometry=[None, None],
        )
        Edge(df=df)


def test_edge_plot(edge):
    edge.plot()



import geopandas as gpd
import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal
from ribasim.geometry.node import Node
from shapely import LineString


def test_utils():
    node_type = ("Basin", "LinearResistance", "Basin")

    xy = np.array([(0.0, 0.0), (1.0, 0.0), (2.0, 0.0)])

    node_xy = gpd.points_from_xy(x=xy[:, 0], y=xy[:, 1])

    node = Node(
        df=gpd.GeoDataFrame(
            data={"node_type": node_type},
            index=pd.Index(np.arange(len(xy)) + 1, name="fid"),
            geometry=node_xy,
            crs="EPSG:28992",
        )
    )

    from_id = np.array([1, 2], dtype=np.int64)
    to_id = np.array([2, 3], dtype=np.int64)

    lines = node.geometry_from_connectivity(from_id, to_id)
    assert lines[0].equals(LineString([xy[from_id[0] - 1], xy[to_id[0] - 1]]))
    assert lines[1].equals(LineString([xy[from_id[1] - 1], xy[to_id[1] - 1]]))

    from_id_, to_id_ = node.connectivity_from_geometry(lines)

    assert_array_equal(from_id, from_id_)
    assert_array_equal(to_id, to_id_)



import re
from abc import ABC, abstractmethod
from collections.abc import Callable, Generator
from contextlib import closing
from contextvars import ContextVar
from pathlib import Path
from sqlite3 import Connection, connect
from typing import (
    Any,
    Generic,
    TypeVar,
    cast,
)

import geopandas as gpd
import pandas as pd
import pandera as pa
from pandera.typing import DataFrame
from pydantic import BaseModel as PydanticBaseModel
from pydantic import (
    ConfigDict,
    DirectoryPath,
    Field,
    PrivateAttr,
    ValidationInfo,
    field_validator,
    model_serializer,
    model_validator,
    validate_call,
)

from ribasim.types import FilePath

__all__ = ("TableModel",)

delimiter = " / "

gpd.options.io_engine = "pyogrio"

context_file_loading: ContextVar[dict[str, Any]] = ContextVar(
    "file_loading", default={}
)

TableT = TypeVar("TableT", bound=pa.DataFrameModel)


def esc_id(identifier: str) -> str:
    """Escape SQLite identifiers."""
    return '"' + identifier.replace('"', '""') + '"'


def exists(connection: Connection, name: str) -> bool:
    """Check if a table exists in a SQLite database."""
    with closing(connection.cursor()) as cursor:
        cursor.execute(
            "SELECT name FROM sqlite_master WHERE type='table' AND name=?", (name,)
        )
        result = cursor.fetchone()
    return result is not None


TABLES = ["profile", "state", "static", "time", "logic", "condition"]


class BaseModel(PydanticBaseModel):
    """Overrides Pydantic BaseModel to set our own config."""

    model_config = ConfigDict(
        validate_assignment=True,
        validate_default=True,
        populate_by_name=True,
        use_enum_values=True,
        extra="allow",
    )

    @classmethod
    def fields(cls) -> list[str]:
        """Return the names of the fields contained in the Model."""
        return list(cls.model_fields.keys())


class FileModel(BaseModel, ABC):
    """Base class to represent models with a file representation.

    It therefore always has a `filepath` and if it is given on
    initialization, it will parse that file.

    This class extends the `model_validator` option of Pydantic,
    so when when a Path is given to a field with type `FileModel`,
    it doesn't error, but actually initializes the `FileModel`.

    Attributes
    ----------
        filepath (Path | None):
            The path of this FileModel.
    """

    filepath: Path | None = Field(default=None, exclude=True, repr=False)

    @model_validator(mode="before")
    @classmethod
    def check_filepath(cls, value: Any) -> Any:
        # Enable initialization with a Path.
        if isinstance(value, dict):
            # Pydantic Model init requires a dict
            filepath = value.get("filepath", None)
            if filepath is not None:
                filepath = Path(filepath)
            data = cls._load(filepath)
            data.update(value)
            return data
        elif isinstance(value, Path | str):
            # Pydantic Model init requires a dict
            data = cls._load(Path(value))
            data["filepath"] = value
            return data
        else:
            return value

    @validate_call
    def set_filepath(self, filepath: Path) -> None:
        """Set the filepath of this instance.

        Args:
            filepath (Path): The filepath to set.
        """
        # Disable assignment validation, which would
        # otherwise trigger check_filepath() and _load() again.
        self.model_config["validate_assignment"] = False
        self.filepath = filepath
        self.model_config["validate_assignment"] = True

    @abstractmethod
    def _save(self, directory: DirectoryPath, input_dir: DirectoryPath) -> None:
        """Save this instance to disk.

        This method needs to be implemented by any class deriving from
        FileModel.
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def _load(cls, filepath: Path | None) -> dict[str, Any]:
        """Load the data at filepath and returns it as a dictionary.

        If a derived FileModel does not load data from disk, this should
        return an empty dictionary.

        Args
        ----
            filepath (Path): Path to the data to load.

        Returns
        -------
            dict: The data stored at filepath
        """
        raise NotImplementedError()


class TableModel(FileModel, Generic[TableT]):
    df: DataFrame[TableT] | None = Field(default=None, exclude=True, repr=False)
    _sort_keys: list[str] = PrivateAttr(default=[])

    @field_validator("df")
    @classmethod
    def check_extra_columns(cls, v: DataFrame[TableT]):
        """Allow only extra columns with `meta_` prefix."""
        if isinstance(v, (pd.DataFrame, gpd.GeoDataFrame)):
            for colname in v.columns:
                if colname == "fid":
                    # Autogenerated on writing, don't carry them
                    v = v.drop(columns=["fid"])  # type: ignore
                elif colname not in cls.columns() and not colname.startswith("meta_"):
                    raise ValueError(
                        f"Unrecognized column '{colname}'. Extra columns need a 'meta_' prefix."
                    )
        return v

    @model_serializer
    def set_model(self) -> str | None:
        return str(self.filepath.name) if self.filepath is not None else None

    @classmethod
    def tablename(cls) -> str:
        """Retrieve tablename based on attached Schema.

        NodeSchema -> Schema
        TabularRatingCurveStaticSchema -> TabularRatingCurve / Static
        """
        names: list[str] = re.sub("([A-Z]+)", r" \1", str(cls.tableschema())).split()
        if len(names) > 2:
            return f"{''.join(names[:-2])}{delimiter}{names[-2].lower()}"
        else:
            return names[0]

    @model_validator(mode="before")
    @classmethod
    def check_dataframe(cls, value: Any) -> Any:
        # Enable initialization with a DataFrame.
        if isinstance(value, pd.DataFrame | gpd.GeoDataFrame):
            value = {"df": value}

        return value

    def node_ids(self) -> set[int]:
        node_ids: set[int] = set()
        if self.df is not None and "node_id" in self.df.columns:
            node_ids.update(self.df["node_id"])

        return node_ids

    @classmethod
    def _load(cls, filepath: Path | None) -> dict[str, Any]:
        db = context_file_loading.get().get("database")
        if filepath is not None:
            adf = cls._from_arrow(filepath)
            # TODO Store filepath?
            return {"df": adf}
        elif db is not None:
            ddf = cls._from_db(db, cls.tablename())
            return {"df": ddf}
        else:
            return {}

    def _save(
        self,
        directory: DirectoryPath,
        input_dir: DirectoryPath,
    ) -> None:
        # TODO directory could be used to save an arrow file
        db_path = context_file_loading.get().get("database")
        if self.df is not None and self.filepath is not None:
            self.sort()
            self._write_arrow(self.filepath, directory, input_dir)
        elif self.df is not None and db_path is not None:
            self.sort()
            self._write_table(db_path)

    def _write_table(self, temp_path: Path) -> None:
        """
        Write the contents of the input to a database.

        Parameters
        ----------
        connection : Connection
            SQLite connection to the database.
        """
        table = self.tablename()
        assert self.df is not None

        # Add `fid` to all tables as primary key
        # Enables editing values manually in QGIS
        df = self.df.copy()
        df["fid"] = range(1, len(df) + 1)

        with closing(connect(temp_path)) as connection:
            df.to_sql(
                table,
                connection,
                index=False,
                if_exists="replace",
                dtype={"fid": "INTEGER PRIMARY KEY AUTOINCREMENT"},
            )

            # Set geopackage attribute table
            with closing(connection.cursor()) as cursor:
                sql = "INSERT INTO gpkg_contents (table_name, data_type, identifier) VALUES (?, ?, ?)"
                cursor.execute(sql, (table, "attributes", table))
            connection.commit()

    def _write_arrow(self, filepath: Path, directory: Path, input_dir: Path) -> None:
        """Write the contents of the input to a an arrow file."""
        assert self.df is not None
        path = directory / input_dir / filepath
        path.parent.mkdir(parents=True, exist_ok=True)
        self.df.to_feather(
            path,
            compression="zstd",
            compression_level=6,
        )

    @classmethod
    def _from_db(cls, path: FilePath, table: str) -> pd.DataFrame | None:
        with connect(path) as connection:
            if exists(connection, table):
                query = f"select * from {esc_id(table)}"
                df = pd.read_sql_query(
                    query, connection, parse_dates={"time": {"format": "ISO8601"}}
                )
            else:
                df = None

            return df

    @classmethod
    def _from_arrow(cls, path: FilePath) -> pd.DataFrame:
        directory = context_file_loading.get().get("directory", Path("."))
        return pd.read_feather(directory / path)

    def sort(self):
        """Sort the table as required.

        Sorting is done automatically before writing the table.
        """
        if self.df is not None:
            self.df.sort_values(self._sort_keys, ignore_index=True, inplace=True)

    @classmethod
    def tableschema(cls) -> TableT:
        """Retrieve Pandera Schema.

        The type of the field `df` is known to always be an DataFrame[TableT]]] | None
        """
        optionalfieldtype = cls.model_fields["df"].annotation
        fieldtype = optionalfieldtype.__args__[0]  # type: ignore
        T: TableT = fieldtype.__args__[0]
        return T

    @classmethod
    def columns(cls) -> list[str]:
        """Retrieve column names."""
        T = cls.tableschema()
        return list(T.to_schema().columns.keys())

    def __repr__(self) -> str:
        # Make sure not to return just "None", because it gets extremely confusing
        # when debugging.
        return f"{self.tablename()}\n{self.df.__repr__()}"

    def _repr_html_(self):
        if self.df is None:
            return self.__repr__()
        else:
            return f"<div>{self.tablename()}</div>" + self.df._repr_html_()


class SpatialTableModel(TableModel[TableT], Generic[TableT]):
    @classmethod
    def _from_db(cls, path: FilePath, table: str):
        with connect(path) as connection:
            if exists(connection, table):
                df = gpd.read_file(path, layer=table, fid_as_index=True)
            else:
                print(f"Can't read from {path}:{table}")
                df = None

            return df

    def _write_table(self, path: FilePath) -> None:
        """
        Write the contents of the input to a database.

        Parameters
        ----------
        path : FilePath
        """

        gdf = gpd.GeoDataFrame(data=self.df)
        gdf = gdf.set_geometry("geometry")
        gdf.index.name = "fid"

        gdf.to_file(path, layer=self.tablename(), driver="GPKG", index=True)

    def sort(self):
        self.df.sort_index(inplace=True)


class ChildModel(BaseModel):
    _parent: Any | None = None
    _parent_field: str | None = None

    @model_validator(mode="after")
    def check_parent(self) -> "ChildModel":
        if self._parent is not None:
            self._parent.model_fields_set.update({self._parent_field})
        return self


class NodeModel(ChildModel):
    """Base class to handle combining the tables for a single node type."""

    @model_serializer(mode="wrap")
    def set_modeld(
        self, serializer: Callable[[type["NodeModel"]], dict[str, Any]]
    ) -> dict[str, Any]:
        content = serializer(self)
        return dict(filter(lambda x: x[1], content.items()))

    @field_validator("*")
    @classmethod
    def set_sort_keys(cls, v: Any, info: ValidationInfo) -> Any:
        """Set sort keys for all TableModels if present in FieldInfo."""
        if isinstance(v, (TableModel,)):
            field = cls.model_fields[getattr(info, "field_name")]
            extra = field.json_schema_extra
            if extra is not None and isinstance(extra, dict):
                # We set sort_keys ourselves as list[str] in json_schema_extra
                # but mypy doesn't know.
                v._sort_keys = cast(list[str], extra.get("sort_keys", []))
        return v

    @classmethod
    def get_input_type(cls):
        return cls.__name__

    @classmethod
    def _layername(cls, field: str) -> str:
        return f"{cls.get_input_type()}{delimiter}{field}"

    def add(*args, **kwargs):
        # TODO This is the new API
        pass

    def tables(self) -> Generator[TableModel[Any], Any, None]:
        for key in self.fields():
            attr = getattr(self, key)
            if isinstance(attr, TableModel):
                yield attr

    def node_ids(self) -> set[int]:
        node_ids: set[int] = set()
        for table in self.tables():
            node_ids.update(table.node_ids())
        return node_ids

    def node_ids_and_types(self) -> tuple[list[int], list[str]]:
        ids = self.node_ids()
        return list(ids), len(ids) * [self.get_input_type()]

    def _save(self, directory: DirectoryPath, input_dir: DirectoryPath, **kwargs):
        for field in self.fields():
            getattr(self, field)._save(
                directory,
                input_dir,
            )

    def _repr_content(self) -> str:
        """Generate a succinct overview of the content.

        Skip "empty" attributes: when the dataframe of a TableModel is None.
        """
        content = []
        for field in self.fields():
            attr = getattr(self, field)
            if isinstance(attr, TableModel):
                if attr.df is not None:
                    content.append(field)
            else:
                content.append(field)
        return ", ".join(content)

    def __repr__(self) -> str:
        content = self._repr_content()
        typename = type(self).__name__
        return f"{typename}({content})"



from enum import Enum

from pydantic import Field

from ribasim.geometry import BasinAreaSchema
from ribasim.input_base import ChildModel, NodeModel, SpatialTableModel, TableModel

# These schemas are autogenerated
from ribasim.schemas import (
    BasinProfileSchema,
    BasinStateSchema,
    BasinStaticSchema,
    BasinSubgridSchema,
    BasinTimeSchema,
    DiscreteControlConditionSchema,
    DiscreteControlLogicSchema,
    FlowBoundaryStaticSchema,
    FlowBoundaryTimeSchema,
    FractionalFlowStaticSchema,
    LevelBoundaryStaticSchema,
    LevelBoundaryTimeSchema,
    LevelDemandStaticSchema,
    LevelDemandTimeSchema,
    LinearResistanceStaticSchema,
    ManningResistanceStaticSchema,
    OutletStaticSchema,
    PidControlStaticSchema,
    PidControlTimeSchema,
    PumpStaticSchema,
    TabulatedRatingCurveStaticSchema,
    TabulatedRatingCurveTimeSchema,
    TerminalStaticSchema,
    UserDemandStaticSchema,
    UserDemandTimeSchema,
)


class Allocation(ChildModel):
    timestep: float | None = None
    use_allocation: bool = False


class Results(ChildModel):
    outstate: str | None = None
    compression: bool = True
    compression_level: int = 6
    subgrid: bool = False


class Solver(ChildModel):
    algorithm: str = "QNDF"
    saveat: float = 86400.0
    dt: float | None = None
    dtmin: float | None = None
    dtmax: float | None = None
    force_dtmin: bool = False
    abstol: float = 1e-06
    reltol: float = 1e-05
    maxiters: int = 1000000000
    sparse: bool = True
    autodiff: bool = True


class Verbosity(str, Enum):
    debug = "debug"
    info = "info"
    warn = "warn"
    error = "error"


class Logging(ChildModel):
    verbosity: Verbosity = Verbosity.info
    timing: bool = False


class Terminal(NodeModel):
    static: TableModel[TerminalStaticSchema] = Field(
        default_factory=TableModel[TerminalStaticSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )


class PidControl(NodeModel):
    static: TableModel[PidControlStaticSchema] = Field(
        default_factory=TableModel[PidControlStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )
    time: TableModel[PidControlTimeSchema] = Field(
        default_factory=TableModel[PidControlTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )


class LevelBoundary(NodeModel):
    static: TableModel[LevelBoundaryStaticSchema] = Field(
        default_factory=TableModel[LevelBoundaryStaticSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    time: TableModel[LevelBoundaryTimeSchema] = Field(
        default_factory=TableModel[LevelBoundaryTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )


class Pump(NodeModel):
    static: TableModel[PumpStaticSchema] = Field(
        default_factory=TableModel[PumpStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )


class TabulatedRatingCurve(NodeModel):
    static: TableModel[TabulatedRatingCurveStaticSchema] = Field(
        default_factory=TableModel[TabulatedRatingCurveStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state", "level"]},
    )
    time: TableModel[TabulatedRatingCurveTimeSchema] = Field(
        default_factory=TableModel[TabulatedRatingCurveTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time", "level"]},
    )


class UserDemand(NodeModel):
    static: TableModel[UserDemandStaticSchema] = Field(
        default_factory=TableModel[UserDemandStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "priority"]},
    )
    time: TableModel[UserDemandTimeSchema] = Field(
        default_factory=TableModel[UserDemandTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "priority", "time"]},
    )


class LevelDemand(NodeModel):
    static: TableModel[LevelDemandStaticSchema] = Field(
        default_factory=TableModel[LevelDemandStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "priority"]},
    )
    time: TableModel[LevelDemandTimeSchema] = Field(
        default_factory=TableModel[LevelDemandTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "priority", "time"]},
    )


class FlowBoundary(NodeModel):
    static: TableModel[FlowBoundaryStaticSchema] = Field(
        default_factory=TableModel[FlowBoundaryStaticSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    time: TableModel[FlowBoundaryTimeSchema] = Field(
        default_factory=TableModel[FlowBoundaryTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )


class Basin(NodeModel):
    profile: TableModel[BasinProfileSchema] = Field(
        default_factory=TableModel[BasinProfileSchema],
        json_schema_extra={"sort_keys": ["node_id", "level"]},
    )
    state: TableModel[BasinStateSchema] = Field(
        default_factory=TableModel[BasinStateSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    static: TableModel[BasinStaticSchema] = Field(
        default_factory=TableModel[BasinStaticSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    time: TableModel[BasinTimeSchema] = Field(
        default_factory=TableModel[BasinTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )
    subgrid: TableModel[BasinSubgridSchema] = Field(
        default_factory=TableModel[BasinSubgridSchema],
        json_schema_extra={"sort_keys": ["subgrid_id", "basin_level"]},
    )
    area: SpatialTableModel[BasinAreaSchema] = Field(
        default_factory=SpatialTableModel[BasinAreaSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )


class ManningResistance(NodeModel):
    static: TableModel[ManningResistanceStaticSchema] = Field(
        default_factory=TableModel[ManningResistanceStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )


class DiscreteControl(NodeModel):
    condition: TableModel[DiscreteControlConditionSchema] = Field(
        default_factory=TableModel[DiscreteControlConditionSchema],
        json_schema_extra={
            "sort_keys": ["node_id", "listen_feature_id", "variable", "greater_than"]
        },
    )
    logic: TableModel[DiscreteControlLogicSchema] = Field(
        default_factory=TableModel[DiscreteControlLogicSchema],
        json_schema_extra={"sort_keys": ["node_id", "truth_state"]},
    )


class Outlet(NodeModel):
    static: TableModel[OutletStaticSchema] = Field(
        default_factory=TableModel[OutletStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )


class LinearResistance(NodeModel):
    static: TableModel[LinearResistanceStaticSchema] = Field(
        default_factory=TableModel[LinearResistanceStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )


class FractionalFlow(NodeModel):
    static: TableModel[FractionalFlowStaticSchema] = Field(
        default_factory=TableModel[FractionalFlowStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )



from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandera as pa
import shapely
from matplotlib.axes import Axes
from numpy.typing import NDArray
from pandera.typing import Series
from pandera.typing.geopandas import GeoSeries

from ribasim.input_base import SpatialTableModel

__all__ = ("Edge",)


class EdgeSchema(pa.SchemaModel):
    name: Series[str] = pa.Field(default="")
    from_node_type: Series[str] = pa.Field(nullable=True)
    from_node_id: Series[int] = pa.Field(default=0, coerce=True)
    to_node_type: Series[str] = pa.Field(nullable=True)
    to_node_id: Series[int] = pa.Field(default=0, coerce=True)
    edge_type: Series[str] = pa.Field(default="flow", coerce=True)
    subnetwork_id: Series[pd.Int64Dtype] = pa.Field(
        default=pd.NA, nullable=True, coerce=True
    )
    geometry: GeoSeries[Any] = pa.Field(default=None, nullable=True)

    class Config:
        add_missing_columns = True


class Edge(SpatialTableModel[EdgeSchema]):
    """
    Defines the connections between nodes.

    Parameters
    ----------
    static : pandas.DataFrame
        Table describing the flow connections.
    """

    def get_where_edge_type(self, edge_type: str) -> NDArray[np.bool_]:
        assert self.df is not None
        return (self.df.edge_type == edge_type).to_numpy()

    def plot(self, **kwargs) -> Axes:
        assert self.df is not None  # Pleases mypy
        kwargs = kwargs.copy()  # Avoid side-effects
        ax = kwargs.get("ax", None)
        color_flow = kwargs.pop("color_flow", None)
        color_control = kwargs.pop("color_control", None)

        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")
            kwargs["ax"] = ax

        kwargs_flow = kwargs.copy()
        kwargs_control = kwargs.copy()

        if color_flow is None:
            color_flow = "#3690c0"  # lightblue
            kwargs_flow["color"] = color_flow
            kwargs_flow["label"] = "Flow edge"
        if color_control is None:
            color_control = "grey"
            kwargs_control["color"] = color_control
            kwargs_control["label"] = "Control edge"

        where_flow = self.get_where_edge_type("flow")
        where_control = self.get_where_edge_type("control")

        if not self.df[where_flow].empty:
            self.df[where_flow].plot(**kwargs_flow)

        if where_control.any():
            self.df[where_control].plot(**kwargs_control)

        # Determine the angle for every caret marker and where to place it.
        coords, index = shapely.get_coordinates(self.df.geometry, return_index=True)
        keep = np.diff(index) == 0
        edge_coords = np.stack((coords[:-1, :], coords[1:, :]), axis=1)[keep]
        x, y = np.mean(edge_coords, axis=1).T
        dx, dy = np.diff(edge_coords, axis=1)[:, 0, :].T
        angle = np.degrees(np.arctan2(dy, dx)) - 90

        # Set the color of the marker to match the line.
        # Black is default, set color_flow otherwise; then set color_control.
        color_index = index[1:][keep]
        color = np.where(where_flow[color_index], color_flow, "k")
        color = np.where(where_control[color_index], color_control, color)

        # A faster alternative may be ax.quiver(). However, getting the scaling
        # right is tedious.
        for m_x, m_y, m_angle, c in zip(x, y, angle, color):
            ax.plot(
                m_x,
                m_y,
                marker=(3, 0, m_angle),
                markersize=5,
                linestyle="None",
                c=c,
            )

        return ax



from typing import Any

import pandera as pa
from pandera.typing import Series
from pandera.typing.geopandas import GeoSeries

from ribasim.schemas import _BaseSchema


class BasinAreaSchema(_BaseSchema):
    node_id: Series[int]
    geometry: GeoSeries[Any] = pa.Field(default=None, nullable=True)



from ribasim.input_base import TableModel
from ribasim.schemas import BasinSubgridSchema


def test_tablemodel_schema():
    schema = TableModel[BasinSubgridSchema].tableschema()
    assert schema == BasinSubgridSchema



import pandas as pd
import pytest
import ribasim
import tomli
from numpy.testing import assert_array_equal
from pandas import DataFrame
from pandas.testing import assert_frame_equal
from pydantic import ValidationError
from ribasim import Pump, Terminal


def __assert_equal(a: DataFrame, b: DataFrame, is_network=False) -> None:
    """Like pandas.testing.assert_frame_equal, but ignoring the index."""
    if a is None and b is None:
        return True

    if is_network:
        # We set this on write, needed for GeoPackage.
        a.index.name = "fid"
        a.index.name = "fid"
    else:
        a = a.reset_index(drop=True)
        b = b.reset_index(drop=True)

    # avoid comparing datetime64[ns] with datetime64[ms]
    if "time" in a:
        a["time"] = a.time.astype("datetime64[ns]")
        b["time"] = b.time.astype("datetime64[ns]")

    if "fid" in a:
        a.drop(columns=["fid"], inplace=True)
    if "fid" in b:
        b.drop(columns=["fid"], inplace=True)

    return assert_frame_equal(a, b)


def test_basic(basic, tmp_path):
    model_orig = basic
    toml_path = tmp_path / "basic/ribasim.toml"
    model_orig.write(toml_path)
    model_loaded = ribasim.Model(filepath=toml_path)

    with open(toml_path, "rb") as f:
        toml_dict = tomli.load(f)

    assert toml_dict["ribasim_version"] == ribasim.__version__

    index_a = model_orig.network.node.df.index.to_numpy(int)
    index_b = model_loaded.network.node.df.index.to_numpy(int)
    assert_array_equal(index_a, index_b)
    __assert_equal(
        model_orig.network.node.df, model_loaded.network.node.df, is_network=True
    )
    __assert_equal(
        model_orig.network.edge.df, model_loaded.network.edge.df, is_network=True
    )
    assert model_loaded.basin.time.df is None


def test_basic_arrow(basic_arrow, tmp_path):
    model_orig = basic_arrow
    model_orig.write(tmp_path / "basic_arrow/ribasim.toml")
    model_loaded = ribasim.Model(filepath=tmp_path / "basic_arrow/ribasim.toml")

    __assert_equal(model_orig.basin.profile.df, model_loaded.basin.profile.df)


def test_basic_transient(basic_transient, tmp_path):
    model_orig = basic_transient
    model_orig.write(tmp_path / "basic_transient/ribasim.toml")
    model_loaded = ribasim.Model(filepath=tmp_path / "basic_transient/ribasim.toml")

    __assert_equal(
        model_orig.network.node.df, model_loaded.network.node.df, is_network=True
    )
    __assert_equal(
        model_orig.network.edge.df, model_loaded.network.edge.df, is_network=True
    )

    time = model_loaded.basin.time
    assert model_orig.basin.time.df.time[0] == time.df.time[0]
    __assert_equal(model_orig.basin.time.df, time.df)
    assert time.df.shape == (1468, 7)


def test_pydantic():
    static_data_bad = pd.DataFrame(data={"node_id": [1, 2, 3]})

    with pytest.raises(ValidationError):
        Pump(static=static_data_bad)


def test_repr():
    static_data = pd.DataFrame(
        data={"node_id": [1, 2, 3], "flow_rate": [1.0, -1.0, 0.0]}
    )

    pump_1 = Pump(static=static_data)
    pump_2 = Pump()

    assert repr(pump_1) == "Pump(static)"
    assert repr(pump_2) == "Pump()"
    # Ensure _repr_html doesn't error
    assert isinstance(pump_1.static._repr_html_(), str)
    assert isinstance(pump_2.static._repr_html_(), str)


def test_extra_columns(basic_transient):
    terminal = Terminal(
        static=pd.DataFrame(
            data={
                "node_id": [1, 2, 3],
                "meta_id": [-1, -2, -3],
            }
        )
    )
    assert "meta_id" in terminal.static.df.columns
    assert (terminal.static.df.meta_id == [-1, -2, -3]).all()

    with pytest.raises(ValidationError):
        # Extra column "extra" needs "meta_" prefix
        Terminal(
            static=pd.DataFrame(
                data={
                    "node_id": [1, 2, 3],
                    "meta_id": [-1, -2, -3],
                    "extra": [-1, -2, -3],
                }
            )
        )

    with pytest.raises(ValidationError):
        # Missing required column "node_id"
        Terminal(
            static=pd.DataFrame(
                data={
                    "meta_id": [-1, -2, -3],
                }
            )
        )


def test_sort(level_setpoint_with_minmax, tmp_path):
    model = level_setpoint_with_minmax
    table = model.discrete_control.condition

    # apply a wrong sort, then call the sort method to restore order
    table.df.sort_values("greater_than", ascending=False, inplace=True)
    assert table.df.iloc[0]["greater_than"] == 15.0
    assert table._sort_keys == [
        "node_id",
        "listen_feature_id",
        "variable",
        "greater_than",
    ]
    table.sort()
    assert table.df.iloc[0]["greater_than"] == 5.0

    # re-apply wrong sort, then check if it gets sorted on write
    table.df.sort_values("greater_than", ascending=False, inplace=True)
    model.write(tmp_path / "basic/ribasim.toml")
    # write sorts the model in place
    assert table.df.iloc[0]["greater_than"] == 5.0
    model_loaded = ribasim.Model(filepath=tmp_path / "basic/ribasim.toml")
    table_loaded = model_loaded.discrete_control.condition
    assert table_loaded.df.iloc[0]["greater_than"] == 5.0
    __assert_equal(table.df, table_loaded.df)


def test_roundtrip(trivial, tmp_path):
    model1 = trivial
    model1dir = tmp_path / "model1"
    model2dir = tmp_path / "model2"
    # read a model and then write it to a different path
    model1.write(model1dir / "ribasim.toml")
    model2 = ribasim.Model(filepath=model1dir / "ribasim.toml")
    model2.write(model2dir / "ribasim.toml")

    assert (model1dir / "database.gpkg").is_file()
    assert (model2dir / "database.gpkg").is_file()

    assert (model1dir / "ribasim.toml").read_text() == (
        model2dir / "ribasim.toml"
    ).read_text()

    # check if all tables are the same
    __assert_equal(model1.network.node.df, model2.network.node.df, is_network=True)
    __assert_equal(model1.network.edge.df, model2.network.edge.df, is_network=True)
    for node1, node2 in zip(model1.nodes().values(), model2.nodes().values()):
        for table1, table2 in zip(node1.tables(), node2.tables()):
            __assert_equal(table1.df, table2.df)



# Automatically generated file. Do not modify.

import pandera as pa
from pandera.dtypes import Timestamp
from pandera.typing import Series


class _BaseSchema(pa.DataFrameModel):
    class Config:
        add_missing_columns = True
        coerce = True


class BasinProfileSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    area: Series[float] = pa.Field(nullable=False)
    level: Series[float] = pa.Field(nullable=False)


class BasinStateSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    level: Series[float] = pa.Field(nullable=False)


class BasinStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    drainage: Series[float] = pa.Field(nullable=True)
    potential_evaporation: Series[float] = pa.Field(nullable=True)
    infiltration: Series[float] = pa.Field(nullable=True)
    precipitation: Series[float] = pa.Field(nullable=True)
    urban_runoff: Series[float] = pa.Field(nullable=True)


class BasinSubgridSchema(_BaseSchema):
    subgrid_id: Series[int] = pa.Field(nullable=False)
    node_id: Series[int] = pa.Field(nullable=False)
    basin_level: Series[float] = pa.Field(nullable=False)
    subgrid_level: Series[float] = pa.Field(nullable=False)


class BasinTimeSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    time: Series[Timestamp] = pa.Field(nullable=False)
    drainage: Series[float] = pa.Field(nullable=True)
    potential_evaporation: Series[float] = pa.Field(nullable=True)
    infiltration: Series[float] = pa.Field(nullable=True)
    precipitation: Series[float] = pa.Field(nullable=True)
    urban_runoff: Series[float] = pa.Field(nullable=True)


class DiscreteControlConditionSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    listen_feature_type: Series[str] = pa.Field(nullable=True)
    listen_feature_id: Series[int] = pa.Field(nullable=False)
    variable: Series[str] = pa.Field(nullable=False)
    greater_than: Series[float] = pa.Field(nullable=False)
    look_ahead: Series[float] = pa.Field(nullable=True)


class DiscreteControlLogicSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    truth_state: Series[str] = pa.Field(nullable=False)
    control_state: Series[str] = pa.Field(nullable=False)


class FlowBoundaryStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    flow_rate: Series[float] = pa.Field(nullable=False)


class FlowBoundaryTimeSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    time: Series[Timestamp] = pa.Field(nullable=False)
    flow_rate: Series[float] = pa.Field(nullable=False)


class FractionalFlowStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    fraction: Series[float] = pa.Field(nullable=False)
    control_state: Series[str] = pa.Field(nullable=True)


class LevelBoundaryStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    level: Series[float] = pa.Field(nullable=False)


class LevelBoundaryTimeSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    time: Series[Timestamp] = pa.Field(nullable=False)
    level: Series[float] = pa.Field(nullable=False)


class LevelDemandStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    min_level: Series[float] = pa.Field(nullable=False)
    max_level: Series[float] = pa.Field(nullable=False)
    priority: Series[int] = pa.Field(nullable=False)


class LevelDemandTimeSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    time: Series[Timestamp] = pa.Field(nullable=False)
    min_level: Series[float] = pa.Field(nullable=False)
    max_level: Series[float] = pa.Field(nullable=False)
    priority: Series[int] = pa.Field(nullable=False)


class LinearResistanceStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    resistance: Series[float] = pa.Field(nullable=False)
    max_flow_rate: Series[float] = pa.Field(nullable=True)
    control_state: Series[str] = pa.Field(nullable=True)


class ManningResistanceStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    length: Series[float] = pa.Field(nullable=False)
    manning_n: Series[float] = pa.Field(nullable=False)
    profile_width: Series[float] = pa.Field(nullable=False)
    profile_slope: Series[float] = pa.Field(nullable=False)
    control_state: Series[str] = pa.Field(nullable=True)


class OutletStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    flow_rate: Series[float] = pa.Field(nullable=False)
    min_flow_rate: Series[float] = pa.Field(nullable=True)
    max_flow_rate: Series[float] = pa.Field(nullable=True)
    min_crest_level: Series[float] = pa.Field(nullable=True)
    control_state: Series[str] = pa.Field(nullable=True)


class PidControlStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    listen_node_type: Series[str] = pa.Field(nullable=True)
    listen_node_id: Series[int] = pa.Field(nullable=False)
    target: Series[float] = pa.Field(nullable=False)
    proportional: Series[float] = pa.Field(nullable=False)
    integral: Series[float] = pa.Field(nullable=False)
    derivative: Series[float] = pa.Field(nullable=False)
    control_state: Series[str] = pa.Field(nullable=True)


class PidControlTimeSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    listen_node_type: Series[str] = pa.Field(nullable=True)
    listen_node_id: Series[int] = pa.Field(nullable=False)
    time: Series[Timestamp] = pa.Field(nullable=False)
    target: Series[float] = pa.Field(nullable=False)
    proportional: Series[float] = pa.Field(nullable=False)
    integral: Series[float] = pa.Field(nullable=False)
    derivative: Series[float] = pa.Field(nullable=False)
    control_state: Series[str] = pa.Field(nullable=True)


class PumpStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    flow_rate: Series[float] = pa.Field(nullable=False)
    min_flow_rate: Series[float] = pa.Field(nullable=True)
    max_flow_rate: Series[float] = pa.Field(nullable=True)
    control_state: Series[str] = pa.Field(nullable=True)


class TabulatedRatingCurveStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    level: Series[float] = pa.Field(nullable=False)
    flow_rate: Series[float] = pa.Field(nullable=False)
    control_state: Series[str] = pa.Field(nullable=True)


class TabulatedRatingCurveTimeSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    time: Series[Timestamp] = pa.Field(nullable=False)
    level: Series[float] = pa.Field(nullable=False)
    flow_rate: Series[float] = pa.Field(nullable=False)


class TerminalStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)


class UserDemandStaticSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    active: Series[pa.BOOL] = pa.Field(nullable=True)
    demand: Series[float] = pa.Field(nullable=False)
    return_factor: Series[float] = pa.Field(nullable=False)
    min_level: Series[float] = pa.Field(nullable=False)
    priority: Series[int] = pa.Field(nullable=False)


class UserDemandTimeSchema(_BaseSchema):
    node_id: Series[int] = pa.Field(nullable=False)
    time: Series[Timestamp] = pa.Field(nullable=False)
    demand: Series[float] = pa.Field(nullable=False)
    return_factor: Series[float] = pa.Field(nullable=False)
    min_level: Series[float] = pa.Field(nullable=False)
    priority: Series[int] = pa.Field(nullable=False)



__version__ = "2024.4.0"


from ribasim.config import (
    Allocation,
    Basin,
    DiscreteControl,
    FlowBoundary,
    FractionalFlow,
    LevelBoundary,
    LevelDemand,
    LinearResistance,
    Logging,
    ManningResistance,
    Outlet,
    PidControl,
    Pump,
    Results,
    Solver,
    TabulatedRatingCurve,
    Terminal,
    UserDemand,
    Verbosity,
)
from ribasim.geometry.edge import Edge, EdgeSchema
from ribasim.geometry.node import Node, NodeSchema
from ribasim.model import Model, Network

__all__ = [
    "Allocation",
    "Basin",
    "DiscreteControl",
    "Edge",
    "EdgeSchema",
    "FlowBoundary",
    "FractionalFlow",
    "LevelBoundary",
    "LevelDemand",
    "LinearResistance",
    "Logging",
    "ManningResistance",
    "Model",
    "Network",
    "Node",
    "NodeSchema",
    "Outlet",
    "PidControl",
    "Pump",
    "Results",
    "Solver",
    "TabulatedRatingCurve",
    "Terminal",
    "UserDemand",
    "Verbosity",
]



import re
from sqlite3 import connect

import pandas as pd
import pytest
from pydantic import ValidationError
from ribasim import Model, Solver
from ribasim.input_base import esc_id
from shapely import Point


def test_repr(basic):
    representation = repr(basic).split("\n")
    assert representation[0] == "ribasim.Model("


def test_solver():
    solver = Solver()
    assert solver.algorithm == "QNDF"  # default
    assert solver.saveat == 86400.0

    solver = Solver(saveat=3600.0)
    assert solver.saveat == 3600.0

    solver = Solver(saveat=float("inf"))
    assert solver.saveat == float("inf")

    solver = Solver(saveat=0)
    assert solver.saveat == 0

    with pytest.raises(ValidationError):
        Solver(saveat="a")


@pytest.mark.xfail(reason="Needs refactor")
def test_invalid_node_type(basic):
    # Add entry with invalid node type
    basic.node.static = basic.node.df._append(
        {"node_type": "InvalidNodeType", "geometry": Point(0, 0)}, ignore_index=True
    )

    with pytest.raises(
        TypeError,
        match=re.escape("Invalid node types detected: [InvalidNodeType].") + ".+",
    ):
        basic.validate_model_node_types()


def test_parent_relationship(basic):
    model = basic
    assert model.pump._parent == model
    assert model.pump._parent_field == "pump"


def test_exclude_unset(basic):
    model = basic
    model.solver.saveat = 86400.0
    d = model.model_dump(exclude_unset=True, exclude_none=True, by_alias=True)
    assert "solver" in d
    assert d["solver"]["saveat"] == 86400.0


def test_invalid_node_id(basic):
    model = basic

    # Add entry with invalid node ID
    df = model.pump.static.df._append(
        {"flow_rate": 1, "node_id": -1, "active": True},
        ignore_index=True,
    )
    # Currently can't handle mixed NaN and None in a DataFrame
    df = df.where(pd.notna(df), None)
    model.pump.static.df = df

    with pytest.raises(
        ValueError,
        match=re.escape("Node IDs must be non-negative integers, got [-1]."),
    ):
        model.validate_model_node_field_ids()


@pytest.mark.xfail(reason="Should be reimplemented by the .add() API.")
def test_node_id_duplicate(basic):
    model = basic

    # Add duplicate node ID
    df = model.pump.static.df._append(
        {"flow_rate": 1, "node_id": 1, "active": True}, ignore_index=True
    )
    # Currently can't handle mixed NaN and None in a DataFrame
    df = df.where(pd.notna(df), None)
    model.pump.static.df = df
    with pytest.raises(
        ValueError,
        match=re.escape("These node IDs were assigned to multiple node types: [1]."),
    ):
        model.validate_model_node_field_ids()


def test_node_ids_misassigned(basic):
    model = basic

    # Misassign node IDs
    model.pump.static.df.loc[0, "node_id"] = 8
    model.fractional_flow.static.df.loc[1, "node_id"] = 7

    with pytest.raises(
        ValueError,
        match="For FractionalFlow, the node IDs in the data tables don't match the node IDs in the network.+",
    ):
        model.validate_model_node_ids()


def test_node_ids_unsequential(basic):
    model = basic

    basin = model.basin
    basin.profile = pd.DataFrame(
        data={
            "node_id": [1, 1, 3, 3, 6, 6, 1000, 1000],
            "area": [0.01, 1000.0] * 4,
            "level": [0.0, 1.0] * 4,
        }
    )
    basin.static.df["node_id"] = [1, 3, 6, 1000]

    model.validate_model_node_field_ids()


def test_tabulated_rating_curve_model(tabulated_rating_curve, tmp_path):
    model_orig = tabulated_rating_curve
    basin_area = tabulated_rating_curve.basin.area.df
    assert basin_area is not None
    assert basin_area.geometry.geom_type[0] == "Polygon"
    model_orig.write(tmp_path / "tabulated_rating_curve/ribasim.toml")
    model_new = Model.read(tmp_path / "tabulated_rating_curve/ribasim.toml")
    pd.testing.assert_series_equal(
        model_orig.tabulated_rating_curve.time.df.time,
        model_new.tabulated_rating_curve.time.df.time,
    )


def test_plot(discrete_control_of_pid_control):
    discrete_control_of_pid_control.plot()


def test_write_adds_fid_in_tables(basic, tmp_path):
    model_orig = basic
    # for node an explicit index was provided
    nrow = len(model_orig.network.node.df)
    assert model_orig.network.node.df.index.name == "fid"
    assert model_orig.network.node.df.index.equals(
        pd.RangeIndex(start=1, stop=nrow + 1)
    )
    # for edge no index was provided, but it still needs to write it to file
    nrow = len(model_orig.network.edge.df)
    assert model_orig.network.edge.df.index.name is None
    assert model_orig.network.edge.df.index.equals(pd.RangeIndex(start=0, stop=nrow))

    model_orig.write(tmp_path / "basic/ribasim.toml")
    with connect(tmp_path / "basic/database.gpkg") as connection:
        query = f"select * from {esc_id('Basin / profile')}"
        df = pd.read_sql_query(query, connection)
        assert "fid" in df.columns
        fids = df["fid"]
        assert fids.equals(pd.Series(range(1, len(fids) + 1)))

        query = "select fid from Node"
        df = pd.read_sql_query(query, connection)
        assert "fid" in df.columns
        fids = df["fid"]
        assert fids.equals(pd.Series(range(1, len(fids) + 1)))

        query = "select fid from Edge"
        df = pd.read_sql_query(query, connection)
        assert "fid" in df.columns
        fids = df["fid"]
        assert fids.equals(pd.Series(range(0, len(fids))))



import datetime
import shutil
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tomli
import tomli_w
from pydantic import (
    DirectoryPath,
    Field,
    field_serializer,
    model_serializer,
    model_validator,
)

import ribasim
from ribasim.config import (
    Allocation,
    Basin,
    DiscreteControl,
    FlowBoundary,
    FractionalFlow,
    LevelBoundary,
    LevelDemand,
    LinearResistance,
    Logging,
    ManningResistance,
    Outlet,
    PidControl,
    Pump,
    Results,
    Solver,
    TabulatedRatingCurve,
    Terminal,
    UserDemand,
)
from ribasim.geometry.edge import Edge
from ribasim.geometry.node import Node
from ribasim.input_base import ChildModel, FileModel, NodeModel, context_file_loading
from ribasim.types import FilePath


class Network(FileModel, NodeModel):
    filepath: Path | None = Field(
        default=Path("database.gpkg"), exclude=True, repr=False
    )

    node: Node = Field(default_factory=Node)
    edge: Edge = Field(default_factory=Edge)

    def n_nodes(self):
        if self.node.df is not None:
            n = len(self.node.df)
        else:
            n = 0

        return n

    @classmethod
    def _load(cls, filepath: Path | None) -> dict[str, Any]:
        directory = context_file_loading.get().get("directory", None)
        if directory is not None:
            context_file_loading.get()["database"] = directory / "database.gpkg"
        return {}

    @classmethod
    def _layername(cls, field: str) -> str:
        return field.capitalize()

    def _save(self, directory, input_dir=Path(".")):
        # We write all tables to a temporary database with a dot prefix,
        # and at the end move this over the target file.
        # This does not throw a PermissionError if the file is open in QGIS.
        directory = Path(directory)
        db_path = directory / input_dir / "database.gpkg"
        db_path = db_path.resolve()
        db_path.parent.mkdir(parents=True, exist_ok=True)
        temp_path = db_path.with_stem(".database")

        # avoid adding tables to existing model
        temp_path.unlink(missing_ok=True)
        context_file_loading.get()["database"] = temp_path

        self.node._save(directory, input_dir)
        self.edge._save(directory, input_dir)

        shutil.move(temp_path, db_path)
        context_file_loading.get()["database"] = db_path

    @model_serializer
    def set_modelname(self) -> str:
        if self.filepath is not None:
            return str(self.filepath.name)
        else:
            return str(self.model_fields["filepath"].default)


class Model(FileModel):
    """
    A full Ribasim model schematisation with all input.

    Ribasim model containing the location of the nodes, the edges between the
    nodes, and the node parametrization.

    Parameters
    ----------
    starttime : datetime.datetime
        Starting time of the simulation.
    endtime : datetime.datetime
        End time of the simulation.

    input_dir: Path = Path(".")
        The directory of the input files.
    results_dir: Path = Path("results")
        The directory of the results files.

    network: Network
        Class containing the topology (nodes and edges) of the model.

    results: Results
        Results configuration options.
    solver: Solver
        Solver configuration options.
    logging: Logging
        Logging configuration options.

    allocation: Allocation
        The allocation configuration.
    basin : Basin
        The waterbodies.
    fractional_flow : FractionalFlow
        Split flows into fractions.
    level_boundary : LevelBoundary
        Boundary condition specifying the water level.
    flow_boundary : FlowBoundary
        Boundary conditions specifying the flow.
    linear_resistance: LinearResistance
        Linear flow resistance.
    manning_resistance : ManningResistance
        Flow resistance based on the Manning formula.
    tabulated_rating_curve : TabulatedRatingCurve
        Tabulated rating curve describing flow based on the upstream water level.
    pump : Pump
        Prescribed flow rate from one basin to the other.
    outlet : Outlet
        Prescribed flow rate from one basin to the other.
    terminal : Terminal
        Water sink without state or properties.
    discrete_control : DiscreteControl
        Discrete control logic.
    pid_control : PidControl
        PID controller attempting to set the level of a basin to a desired value using a pump/outlet.
    user_demand : UserDemand
        UserDemand node type with demand and priority.
    """

    starttime: datetime.datetime
    endtime: datetime.datetime

    input_dir: Path = Field(default_factory=lambda: Path("."))
    results_dir: Path = Field(default_factory=lambda: Path("results"))

    network: Network = Field(default_factory=Network, alias="database", exclude=True)
    results: Results = Results()
    solver: Solver = Solver()
    logging: Logging = Logging()

    allocation: Allocation = Field(default_factory=Allocation)
    level_demand: LevelDemand = Field(default_factory=LevelDemand)
    basin: Basin = Field(default_factory=Basin)
    fractional_flow: FractionalFlow = Field(default_factory=FractionalFlow)
    level_boundary: LevelBoundary = Field(default_factory=LevelBoundary)
    flow_boundary: FlowBoundary = Field(default_factory=FlowBoundary)
    linear_resistance: LinearResistance = Field(default_factory=LinearResistance)
    manning_resistance: ManningResistance = Field(default_factory=ManningResistance)
    tabulated_rating_curve: TabulatedRatingCurve = Field(
        default_factory=TabulatedRatingCurve
    )
    pump: Pump = Field(default_factory=Pump)
    outlet: Outlet = Field(default_factory=Outlet)
    terminal: Terminal = Field(default_factory=Terminal)
    discrete_control: DiscreteControl = Field(default_factory=DiscreteControl)
    pid_control: PidControl = Field(default_factory=PidControl)
    user_demand: UserDemand = Field(default_factory=UserDemand)

    @model_validator(mode="after")
    def set_node_parent(self) -> "Model":
        for (
            k,
            v,
        ) in self.children().items():
            setattr(v, "_parent", self)
            setattr(v, "_parent_field", k)
        return self

    @field_serializer("input_dir", "results_dir")
    def serialize_path(self, path: Path) -> str:
        return str(path)

    def model_post_init(self, __context: Any) -> None:
        # Always write dir fields
        self.model_fields_set.update({"input_dir", "results_dir"})

    def __repr__(self) -> str:
        """Generate a succinct overview of the Model content.

        Skip "empty" NodeModel instances: when all dataframes are None.
        """
        content = ["ribasim.Model("]
        INDENT = "    "
        for field in self.fields():
            attr = getattr(self, field)
            if isinstance(attr, NodeModel):
                attr_content = attr._repr_content()
                typename = type(attr).__name__
                if attr_content:
                    content.append(f"{INDENT}{field}={typename}({attr_content}),")
            else:
                content.append(f"{INDENT}{field}={repr(attr)},")

        content.append(")")
        return "\n".join(content)

    def _write_toml(self, fn: FilePath):
        fn = Path(fn)

        content = self.model_dump(exclude_unset=True, exclude_none=True, by_alias=True)
        # Filter empty dicts (default Nodes)
        content = dict(filter(lambda x: x[1], content.items()))
        content["ribasim_version"] = ribasim.__version__
        with open(fn, "wb") as f:
            tomli_w.dump(content, f)
        return fn

    def _save(self, directory: DirectoryPath, input_dir: DirectoryPath):
        self.network._save(directory, input_dir)
        for sub in self.nodes().values():
            sub._save(directory, input_dir)

    def nodes(self):
        return {
            k: getattr(self, k)
            for k in self.model_fields.keys()
            if isinstance(getattr(self, k), NodeModel) and k != "network"
        }

    def children(self):
        return {
            k: getattr(self, k)
            for k in self.model_fields.keys()
            if isinstance(getattr(self, k), ChildModel)
        }

    def validate_model_node_field_ids(self):
        """Check whether the node IDs of the node_type fields are valid."""

        # Check node IDs of node fields
        all_node_ids = set[int]()
        for node in self.nodes().values():
            all_node_ids.update(node.node_ids())

        unique, counts = np.unique(list(all_node_ids), return_counts=True)
        node_ids_negative_integers = np.less(unique, 0) | np.not_equal(
            unique.astype(np.int64), unique
        )

        if node_ids_negative_integers.any():
            raise ValueError(
                f"Node IDs must be non-negative integers, got {unique[node_ids_negative_integers]}."
            )

        if (counts > 1).any():
            raise ValueError(
                f"These node IDs were assigned to multiple node types: {unique[(counts > 1)]}."
            )

    def validate_model_node_ids(self):
        """Check whether the node IDs in the data tables correspond to the node IDs in the network."""

        error_messages = []

        for node in self.nodes().values():
            nodetype = node.get_input_type()
            node_ids_data = set(node.node_ids())
            node_ids_network = set(
                self.network.node.df.loc[
                    self.network.node.df["node_type"] == nodetype
                ].index
            )

            if not node_ids_network == node_ids_data:
                extra_in_network = node_ids_network.difference(node_ids_data)
                extra_in_data = node_ids_data.difference(node_ids_network)
                error_messages.append(
                    f"""For {nodetype}, the node IDs in the data tables don't match the node IDs in the network.
    Node IDs only in the data tables: {extra_in_data}.
    Node IDs only in the network: {extra_in_network}.
                    """
                )

        if len(error_messages) > 0:
            raise ValueError("\n".join(error_messages))

    def validate_model(self):
        """Validate the model.

        Checks:
        - Whether the node IDs of the node_type fields are valid
        - Whether the node IDs in the node field correspond to the node IDs on the node type fields
        """

        self.validate_model_node_field_ids()
        self.validate_model_node_ids()

    def _add_node_type(self, df: pd.DataFrame | None, id_col: str, type_col: str):
        node = self.network.node.df
        assert node is not None
        if df is not None:
            df[type_col] = node.loc[df[id_col], "node_type"].to_numpy()

    def _add_node_types(self):
        """Add the from/to node types to tables that reference external node IDs.

        Only valid with globally unique node IDs, which is assured by using the node index.
        """
        self._add_node_type(self.network.edge.df, "from_node_id", "from_node_type")
        self._add_node_type(self.network.edge.df, "to_node_id", "to_node_type")
        id_col, type_col = "listen_node_id", "listen_node_type"
        self._add_node_type(self.pid_control.static.df, id_col, type_col)
        self._add_node_type(self.pid_control.time.df, id_col, type_col)
        self._add_node_type(
            self.discrete_control.condition.df,
            "listen_feature_id",
            "listen_feature_type",
        )

    @classmethod
    def read(cls, filepath: FilePath) -> "Model":
        """Read model from TOML file."""
        return cls(filepath=filepath)  # type: ignore

    def write(self, filepath: Path | str) -> Path:
        """
        Write the contents of the model to disk and save it as a TOML configuration file.

        If ``filepath.parent`` does not exist, it is created before writing.

        Parameters
        ----------
        filepath: FilePath ending in .toml
        """
        self.validate_model()
        self._add_node_types()
        filepath = Path(filepath)
        if not filepath.suffix == ".toml":
            raise ValueError(f"Filepath '{filepath}' is not a .toml file.")
        context_file_loading.set({})
        filepath = Path(filepath)
        directory = filepath.parent
        directory.mkdir(parents=True, exist_ok=True)
        self._save(directory, self.input_dir)
        fn = self._write_toml(filepath)

        context_file_loading.set({})
        return fn

    @classmethod
    def _load(cls, filepath: Path | None) -> dict[str, Any]:
        context_file_loading.set({})

        if filepath is not None:
            with open(filepath, "rb") as f:
                config = tomli.load(f)

            context_file_loading.get()["directory"] = filepath.parent / config.get(
                "input_dir", "."
            )
            return config
        else:
            return {}

    @model_validator(mode="after")
    def reset_contextvar(self) -> "Model":
        # Drop database info
        context_file_loading.set({})
        return self

    def plot_control_listen(self, ax):
        x_start, x_end = [], []
        y_start, y_end = [], []

        condition = self.discrete_control.condition.df
        if condition is not None:
            for node_id in condition.node_id.unique():
                data_node_id = condition[condition.node_id == node_id]

                for listen_feature_id in data_node_id.listen_feature_id:
                    point_start = self.network.node.df.iloc[node_id - 1].geometry
                    x_start.append(point_start.x)
                    y_start.append(point_start.y)

                    point_end = self.network.node.df.iloc[
                        listen_feature_id - 1
                    ].geometry
                    x_end.append(point_end.x)
                    y_end.append(point_end.y)

        for table in [self.pid_control.static.df, self.pid_control.time.df]:
            if table is None:
                continue

            node = self.network.node.df

            for node_id in table.node_id.unique():
                for listen_node_id in table.loc[
                    table.node_id == node_id, "listen_node_id"
                ].unique():
                    point_start = node.iloc[listen_node_id - 1].geometry
                    x_start.append(point_start.x)
                    y_start.append(point_start.y)

                    point_end = node.iloc[node_id - 1].geometry
                    x_end.append(point_end.x)
                    y_end.append(point_end.y)

        if len(x_start) == 0:
            return

        # This part can probably be done more efficiently
        for i, (x, y, x_, y_) in enumerate(zip(x_start, y_start, x_end, y_end)):
            ax.plot(
                [x, x_],
                [y, y_],
                c="gray",
                ls="--",
                label="Listen edge" if i == 0 else None,
            )

    def plot(self, ax=None, indicate_subnetworks: bool = True) -> Any:
        """
        Plot the nodes, edges and allocation networks of the model.

        Parameters
        ----------
        ax : matplotlib.pyplot.Artist, optional
            Axes on which to draw the plot.

        Returns
        -------
        ax : matplotlib.pyplot.Artist
        """
        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")

        self.network.edge.plot(ax=ax, zorder=2)
        self.plot_control_listen(ax)
        self.network.node.plot(ax=ax, zorder=3)

        handles, labels = ax.get_legend_handles_labels()

        if indicate_subnetworks:
            (
                handles_subnetworks,
                labels_subnetworks,
            ) = self.network.node.plot_allocation_networks(ax=ax, zorder=1)
            handles += handles_subnetworks
            labels += labels_subnetworks

        ax.legend(handles, labels, loc="lower left", bbox_to_anchor=(1, 0.5))

        return ax

    def print_discrete_control_record(self, path: FilePath) -> None:
        path = Path(path)
        df_control = pd.read_feather(path)
        node_attrs, node_instances = zip(*self.nodes().items())
        node_clss = [node_cls.get_input_type() for node_cls in node_instances]
        truth_dict = {"T": ">", "F": "<"}
        assert self.network.node.df is not None
        assert self.network.edge.df is not None

        if self.discrete_control.condition.df is None:
            raise ValueError("This model has no control input.")

        for index, row in df_control.iterrows():
            datetime = row["time"]
            control_node_id = row["control_node_id"]
            truth_state = row["truth_state"]
            control_state = row["control_state"]
            enumeration = f"{index}. "

            out = f"{enumeration}At {datetime} the control node with ID {control_node_id} reached truth state {truth_state}:\n"

            if self.discrete_control.condition.df is None:
                return

            conditions = self.discrete_control.condition.df[
                self.discrete_control.condition.df.node_id == control_node_id
            ]

            for truth_value, (index, condition) in zip(
                truth_state, conditions.iterrows()
            ):
                var = condition["variable"]
                listen_feature_id = condition["listen_feature_id"]
                listen_node_type = self.network.node.df.loc[
                    listen_feature_id, "node_type"
                ]
                symbol = truth_dict[truth_value]
                greater_than = condition["greater_than"]
                feature_type = "edge" if var == "flow" else "node"

                out += f"\tFor {feature_type} ID {listen_feature_id} ({listen_node_type}): {var} {symbol} {greater_than}\n"

            padding = len(enumeration) * " "
            out += f'\n{padding}This yielded control state "{control_state}":\n'

            affect_node_ids = self.network.edge.df[
                self.network.edge.df.from_node_id == control_node_id
            ].to_node_id

            for affect_node_id in affect_node_ids:
                affect_node_type = self.network.node.df.loc[affect_node_id, "node_type"]
                nodeattr = node_attrs[node_clss.index(affect_node_type)]

                out += f"\tFor node ID {affect_node_id} ({affect_node_type}): "

                static = getattr(self, nodeattr).static.df
                row = static[
                    (static.node_id == affect_node_id)
                    & (static.control_state == control_state)
                ].iloc[0]

                names_and_values = []
                for var in static.columns:
                    if var not in ["node_id", "control_state"]:
                        value = row[var]
                        if value is not None:
                            names_and_values.append(f"{var} = {value}")

                out += ", ".join(names_and_values) + "\n"

            print(out)
